# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================
import math
from typing import Any

from libinephany.observations.observation_utils import StatisticStorageTypes
from libinephany.observations.observers.base_observers import GlobalObserver
from libinephany.observations.observers.global_observers.base_classes import LHOPTCheckpointBaseObserver
from libinephany.pydantic_models.schemas.observation_models import ObservationInputs
from libinephany.pydantic_models.schemas.tensor_statistics import TensorStatistics
from libinephany.pydantic_models.states.hyperparameter_states import HyperparameterStates

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class TrainingProgress(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.training_progress

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class EpochsCompleted(GlobalObserver):

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """

        return observation_inputs.epochs_completed

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}


class ProgressAtEachCheckpoint(LHOPTCheckpointBaseObserver):
    """
    This is a global observer from the paper "Learning to Optimize with Reinforcement Learning"
    https://arxiv.org/abs/2305.18291.

    It returns a single float value that is the training progress at the current checkpoint.
    The observation tracks training progress and returns the progress value only when a checkpoint is reached.
    """

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float:
        """
        Returns training progress at each checkpoint interval.

        Before checkpoint: returns progress towards next checkpoint (0 to 1)
        At checkpoint: returns the actual training progress value
        """
        current_progress = observation_inputs.training_progress

        # Cold start: If the last progress is not set, set it to the first progress record
        self._cold_start(current_progress)

        self._update_history(current_progress)

        # Check if we should create a checkpoint
        if self._should_create_checkpoint():
            # Return the progress at this checkpoint
            self.last_value = current_progress
            return current_progress
        else:
            return self.last_value


class StagnationObserver(GlobalObserver):

    def __init__(
        self, use_validation_loss: bool, stagnation_threshold: float = 0.0, log_transform: bool = True, **kwargs
    ):
        """
        :param stagnation_threshold: The loss improvement threshold. Default is 0.0. This is the amount the training
        loss must improve by over the best observed loss (training or validation, depends on the use_validation_loss parameter) to reset the stagnation counter.
        We assume that improvement means loss decreases regardless of the sign of the loss, which means the best loss is
        negative infinity.
        :param log_transform: Whether to log transform the stagnation counter. Default is True.
        :param kwargs: Miscellaneous keyword arguments.
        """
        super().__init__(**kwargs)
        self.use_validation_loss = use_validation_loss
        self.best_loss: float | None = None
        self.stagnation_counter: int = 0
        self.stagnation_threshold = stagnation_threshold
        self.log_transform = log_transform

    @property
    def requires_validation_loss(self) -> bool:
        """
        :return: Whether the observation requires validation loss to be calculated.
        """

        return self.use_validation_loss

    @property
    def can_standardize(self) -> bool:
        """
        :return: Whether the observation can be standardized.
        """

        return False

    def _get_observation_format(self) -> StatisticStorageTypes:
        """
        :return: Format the observation returns data in. Must be one of the enum attributes in the StatisticStorageTypes
        enumeration class.
        """

        return StatisticStorageTypes.FLOAT

    def _get_loss(self, observation_inputs: ObservationInputs) -> float:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :return: The loss to use for the stagnation observer.
        """
        return observation_inputs.validation_loss if self.requires_validation_loss else observation_inputs.training_loss

    def _observe(
        self,
        observation_inputs: ObservationInputs,
        hyperparameter_states: HyperparameterStates,
        tracked_statistics: dict[str, dict[str, float | TensorStatistics]],
        action_taken: float | int | None,
        num_categories: int | None,
    ) -> float | int | list[int | float] | TensorStatistics:
        """
        :param observation_inputs: Observation input metrics not calculated with statistic trackers.
        :param hyperparameter_states: HyperparameterStates that manages the hyperparameters.
        :param tracked_statistics: Dictionary mapping statistic tracker class names to dictionaries mapping module
        names to floats or TensorStatistic models.
        :param action_taken: Action taken by the agent this class instance is assigned to.
        :param num_categories: Number of categories, used to normalised loss based observations.
        :return: Single float/int, list of floats/ints or TensorStatistics model to add to the observation vector.
        """
        # Validate input values, if loss is invalid, increment stagnation counter
        if math.isnan(self._get_loss(observation_inputs)) or math.isinf(self._get_loss(observation_inputs)):
            self.stagnation_counter += 1
        else:
            if self.best_loss is None:
                self.best_loss = self._get_loss(observation_inputs)
                self.stagnation_counter = 0
            else:
                improvement = self.best_loss - self._get_loss(observation_inputs)
                if improvement > self.stagnation_threshold:
                    self.best_loss = self._get_loss(observation_inputs)
                    self.stagnation_counter = 0
                else:
                    self.stagnation_counter += 1
        return self.stagnation_counter if not self.log_transform else math.log(self.stagnation_counter + 1)

    def get_required_trackers(self) -> dict[str, dict[str, Any] | None]:
        """
        :return: Dictionary mapping statistic tracker class names to kwargs for the class or None if no kwargs are
        needed.
        """

        return {}



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from typing import Any, cast

from pydantic import BaseModel, ConfigDict, ValidationError, field_serializer, field_validator, model_validator

from libinephany.utils.enums import AgentTypes

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

SKIP_TYPE_CONVERSIONS = ["initial_delta", "scale", "dynamics_kwargs", "sample_mean", "sample_sigma"]

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class HParamConfig(BaseModel):

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    max_hparam_value: float | int
    min_hparam_value: float | int
    initial_value: float | int
    initial_delta: float
    scale: float
    hparam_dtype: type[float | int] = float
    transform: str = "identity"

    sampler: str
    sample_initial_values: bool = False
    sample_lower_bound: float | int
    sample_upper_bound: float | int
    sample_step: int | float | None = None
    sample_discrete_values: list[float | int] | None = None
    sample_mean: int | float | None = None
    sample_sigma: int | float | None = None

    dynamics_function_name: str | None = None
    dynamics_kwargs: dict[str, float | int | str] = {}
    clipping_zeros_velocity: bool = False

    @field_validator("hparam_dtype", mode="before")
    @classmethod
    def _parse_dtype(cls, dtype: type[float | int] | str | None) -> type[float | int]:
        if dtype is None:
            return float
        if type(dtype) is type:
            return dtype
        if type(dtype) is str:
            if dtype == "float":
                return float
            if dtype == "int":
                return int
        raise ValidationError

    @model_validator(mode="after")
    def _ensure_type(self) -> "HParamConfig":
        """
        :return: Updated hparam config.
        """

        for field, value in self.__dict__.items():
            if field in SKIP_TYPE_CONVERSIONS:
                continue

            if value is not None and not isinstance(value, (str, type, bool, list)):
                self.__dict__[field] = self.hparam_dtype(value)

        return self

    @field_serializer("hparam_dtype")
    def serialize_type(self, value: type[int | float], *args, **kwargs) -> str:
        """
        :param value: Value to serialize.
        :return: Serialized type.
        """

        return value.__name__

    def override(self, overrides: dict[str, Any]) -> None:
        """
        :param overrides: Overrides to apply to the hyperparameter config.
        """

        for field, value in overrides.items():
            if not hasattr(self, field):
                raise AttributeError(f"{self.__class__.__name__} does not have field {field} to override!")

            if not isinstance(field, (str, type, bool)):
                setattr(self, field, self.hparam_dtype(value))

            else:
                setattr(self, field, value)

        self.model_validate(self, strict=True)


class LearningRateHParamConfig(HParamConfig):
    max_hparam_value: float | int = 1e-2
    min_hparam_value: float | int = 1e-10
    initial_value: float = 0.001
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "RoundRobinDiscreteValueSampler"
    sample_initial_values: bool = True
    sample_lower_bound: float | int = 1e-5
    sample_upper_bound: float | int = 1e-2
    sample_discrete_values: list[float | int] | None = [0.0, 0.01, 0.001, 0.0001, 0.003, 0.0003]


class WeightDecayHParamConfig(HParamConfig):
    max_hparam_value: float | int = 1.0
    min_hparam_value: float | int = 0.0
    initial_value: float = 0.001
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "RoundRobinDiscreteValueSampler"
    sample_initial_values: bool = True
    sample_lower_bound: float | int = 1e-5
    sample_upper_bound: float | int = 1e-1
    sample_discrete_values: list[float | int] | None = [0.0, 0.1, 0.01, 0.001, 0.0001, 0.00001]


class DropoutHParamConfig(HParamConfig):
    max_hparam_value: float | int = 0.9
    min_hparam_value: float | int = 0.0
    initial_value: float = 0.05
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: float | int = 0.0
    sample_upper_bound: float | int = 0.5
    sample_step: float = 0.1


class GradientNormClippingHParamConfig(HParamConfig):
    max_hparam_value: float | int = 10.0
    min_hparam_value: float | int = 0.1
    initial_value: float = 1.0
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "RoundRobinDiscreteValueSampler"
    sample_initial_values: bool = False
    sample_lower_bound: float | int = 5.0
    sample_upper_bound: float | int = 1e-1
    sample_discrete_values: list[float | int] | None = [0.1, 0.5, 1.0, 3.0, 5.0, 10.0]


class AdamBetaOneHParamConfig(HParamConfig):
    max_hparam_value: float | int = 0.99
    min_hparam_value: float | int = 0.0
    initial_value: float = 0.9
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: float = 0.0
    sample_upper_bound: float = 0.99
    sample_step: float = 0.01
    sample_discrete_values: list[float | int] | None = None


class AdamBetaTwoHParamConfig(HParamConfig):
    max_hparam_value: float | int = 0.99
    min_hparam_value: float | int = 0.0
    initial_value: float = 0.99
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: float = 0.0
    sample_upper_bound: float = 0.99
    sample_step: float = 0.01
    sample_discrete_values: list[float | int] | None = None


class AdamBetaGainHParamConfig(HParamConfig):
    max_hparam_value: float | int = 1.0
    min_hparam_value: float | int = 0.0
    initial_value: float = 0.9
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: float | int = 0.0
    sample_upper_bound: float | int = 0.99
    sample_step: float = 0.1


class AdamEpsHParamConfig(HParamConfig):
    max_hparam_value: float | int = 1e-3
    min_hparam_value: float | int = 1e-15
    initial_value: float = 1e-8
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: float = 0.0
    sample_upper_bound: float = 0.99
    sample_step: float = 0.01
    sample_discrete_values: list[float | int] | None = None


class BatchSizeHParamConfig(HParamConfig):
    max_hparam_value: float | int = 512
    min_hparam_value: float | int = 8
    hparam_dtype: type[float | int] = int
    initial_value: int = 128
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: int = 8
    sample_upper_bound: int = 512
    sample_step: int = 8
    sample_discrete_values: list[float | int] | None = None


class GradientAccumulationHParamConfig(HParamConfig):
    max_hparam_value: float | int = 64
    min_hparam_value: float | int = 1
    hparam_dtype: type[float | int] = int
    initial_value: int = 1
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: int = 1
    sample_upper_bound: int = 64
    sample_step: int = 1
    sample_discrete_values: list[float | int] | None = None
    force_limit: float | int = 64


class EpochsHParamConfig(HParamConfig):
    max_hparam_value: float | int = 16
    min_hparam_value: float | int = 1
    hparam_dtype: type[float | int] = int
    initial_value: int = 1
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: int = 1
    sample_upper_bound: int = 3
    sample_step: int = 1


class TokensHParamConfig(HParamConfig):
    max_hparam_value: float | int = 1e9
    min_hparam_value: float | int = 25e6
    hparam_dtype: type[float | int] = int
    initial_value: int = int(25e6)
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: int = int(25e6)
    sample_upper_bound: int = int(1e9)
    sample_step: int = int(1e6)


class SamplesHParamConfig(HParamConfig):
    max_hparam_value: float | int = 1e9
    min_hparam_value: float | int = 1e4
    hparam_dtype: type[float | int] = int
    initial_value: int = int(1e5)
    initial_delta: float = 0.0
    scale: float = 1.0

    sampler: str = "DiscreteRangeSampler"
    sample_initial_values: bool = False
    sample_lower_bound: int = int(1e4)
    sample_upper_bound: int = int(1e7)
    sample_step: int = int(1e4)


class HParamConfigs(BaseModel):

    learning_rate_config: HParamConfig = LearningRateHParamConfig()
    weight_decay_config: HParamConfig = WeightDecayHParamConfig()
    dropout_config: HParamConfig = DropoutHParamConfig()
    gradient_norm_clipping_config: HParamConfig = GradientNormClippingHParamConfig()
    adam_beta_one_config: HParamConfig = AdamBetaOneHParamConfig()
    adam_beta_two_config: HParamConfig = AdamBetaTwoHParamConfig()
    adam_beta_gain_config: HParamConfig = AdamBetaGainHParamConfig()
    adam_eps_config: HParamConfig = AdamEpsHParamConfig()

    batch_size_config: HParamConfig = BatchSizeHParamConfig()
    gradient_accumulation_config: GradientAccumulationHParamConfig = GradientAccumulationHParamConfig()
    epochs_config: HParamConfig = EpochsHParamConfig()
    token_config: HParamConfig = TokensHParamConfig()
    samples_config: HParamConfig = SamplesHParamConfig()

    constraints: dict[str, dict[str, Any] | None] | None = {
        "FirstDefaultParamorphConstraint": None,
    }

    def override(self, overrides: dict[str, dict[str, Any]] | None) -> None:
        """
        :param overrides: None or config overrides to apply.
        """

        if overrides is not None:
            for config_name, config_overrides in overrides.items():
                if not hasattr(self, config_name):
                    raise AttributeError(
                        f"{self.__class__.__name__} does not have hyperparameter config {config_name}!"
                    )

                config = getattr(self, config_name)
                config.override(overrides=config_overrides)

    def set_hyperparameter_config_from_agent_type(
        self, agent_type: str | AgentTypes, hparam_config: HParamConfig
    ) -> None:
        """
        :param agent_type: Type of agent to set the hyperparameter config for.
        :param hparam_config: HParamConfig for the agent type to set.
        :raises ValueError: If given agent type is not known.
        """

        agent_type = AgentTypes(agent_type)

        match agent_type:
            case AgentTypes.LearningRateAgent:
                self.learning_rate_config = hparam_config

            case AgentTypes.WeightDecayAgent:
                self.weight_decay_config = hparam_config

            case AgentTypes.DropoutAgent:
                self.dropout_config = hparam_config

            case AgentTypes.GradientClippingAgent:
                self.gradient_norm_clipping_config = hparam_config

            case AgentTypes.AdamBetaOneAgent:
                self.adam_beta_one_config = hparam_config

            case AgentTypes.AdamBetaTwoAgent:
                self.adam_beta_two_config = hparam_config

            case AgentTypes.AdamBetaGainAgent:
                self.adam_beta_gain_config = hparam_config

            case AgentTypes.AdamEpsAgent:
                self.adam_eps_config = hparam_config

            case AgentTypes.BatchSize:
                self.batch_size_config = hparam_config

            case AgentTypes.GradientAccumulationAgent:
                self.gradient_accumulation_config = cast(GradientAccumulationHParamConfig, hparam_config)

            case AgentTypes.Epochs:
                self.epochs_config = hparam_config

            case AgentTypes.Tokens:
                self.token_config = hparam_config

            case AgentTypes.Samples:
                self.samples_config = hparam_config

            case _:
                raise ValueError(f"Unknown agent type: {agent_type.value}.")

    def get_hyperparameter_config_from_agent_type(self, agent_type: str | AgentTypes) -> HParamConfig:
        """
        :param agent_type: Type of agent to get the hyperparameter config for.
        :return: Corresponding hyperparameter config to the given agent type.
        :raises ValueError: If given agent type is not known.
        """

        agent_type = AgentTypes(agent_type)

        match agent_type:
            case AgentTypes.LearningRateAgent:
                return self.learning_rate_config

            case AgentTypes.WeightDecayAgent:
                return self.weight_decay_config

            case AgentTypes.DropoutAgent:
                return self.dropout_config

            case AgentTypes.GradientClippingAgent:
                return self.gradient_norm_clipping_config

            case AgentTypes.AdamBetaOneAgent:
                return self.adam_beta_one_config

            case AgentTypes.AdamBetaTwoAgent:
                return self.adam_beta_two_config

            case AgentTypes.AdamBetaGainAgent:
                return self.adam_beta_gain_config

            case AgentTypes.AdamEpsAgent:
                return self.adam_eps_config

            case AgentTypes.BatchSize:
                return self.batch_size_config

            case AgentTypes.GradientAccumulationAgent:
                return self.gradient_accumulation_config

            case AgentTypes.Epochs:
                return self.epochs_config

            case AgentTypes.Tokens:
                return self.token_config

            case AgentTypes.Samples:
                return self.samples_config

            case _:
                raise ValueError(f"Unknown agent type: {agent_type.value}.")



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from functools import partial
from typing import Callable

import torch
from pydantic import BaseModel, Field

# ======================================================================================================================
#
# CONSTANTS
#
# ======================================================================================================================

DEFAULT_TRANSFORM = "identity"
LOG_MIN_VALUE = -100.0  # log of bf16/f32 smallest value is -87.7
TRANSFORM_REGISTRY: dict[str, Callable[[torch.Tensor | float], float]] = {
    "identity": lambda x: torch.as_tensor(x).item(),
    "log": lambda x: max(torch.log(torch.as_tensor(x)).item(), LOG_MIN_VALUE),
    "abs": lambda x: torch.abs(torch.as_tensor(x)).item(),
    "1mexp": lambda x: -torch.expm1(torch.as_tensor(x)).item(),  # 1 - exp(x)
    "squared": lambda x: (torch.as_tensor(x) ** 2).item(),
}
FROM_TENSOR_FN = "from_tensor_fn"


# ======================================================================================================================
#
# TYPE ALIASES
#
# ======================================================================================================================


IncludeStatisticsType = dict[str, str | None] | None


# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class TensorStatistics(BaseModel):

    norm: float = Field(0.0, from_tensor_fn=partial(torch.norm, p=2))  # type: ignore
    mean: float = Field(0.0, from_tensor_fn=torch.mean)  # type: ignore
    median: float = Field(0.0, from_tensor_fn=torch.median)  # type: ignore
    variance: float = Field(0.0, from_tensor_fn=partial(torch.var, unbiased=False))  # type: ignore
    tenth_percentile: float = Field(0.0, from_tensor_fn=partial(torch.quantile, q=0.1))  # type: ignore
    ninetieth_percentile: float = Field(0.0, from_tensor_fn=partial(torch.quantile, q=0.9))  # type: ignore
    inter_quartile_range: float = Field(
        0.0, from_tensor_fn=lambda tensor: torch.quantile(tensor, 0.75) - torch.quantile(tensor, 0.25)  # type: ignore
    )

    @staticmethod
    def downsample_tensor(tensor: torch.Tensor, sample_percentage: float) -> torch.Tensor:
        """
        :param tensor: Tensor to downsample.
        :param sample_percentage: Percentage of the given tensor to randomly sample and compute statistics from.
        :return: Downsampled tensor.
        """

        if sample_percentage >= 1.0:
            return tensor

        input_size = tensor.numel()
        sample_size = max(int(input_size * sample_percentage), 1)

        random_indices = torch.randint(0, input_size, (sample_size,), device=tensor.device)
        tensor = tensor.view(-1)

        return tensor[random_indices]

    @classmethod
    def build(
        cls,
        tensor: torch.Tensor,
        include_statistics: IncludeStatisticsType,
        sample_percentage: float = 1.0,
    ) -> "TensorStatistics":
        """
        :param tensor: Tensor to compute and store statistics of.
        :param include_statistics: If the observation uses the TensorStatistic model to return observations, names of the
        fields in the model to include in returned observations, optionally with transform names.
        :param sample_percentage: Percentage of the given tensor to randomly sample and compute statistics from.
        :return: Constructed tensor statistics.
        """

        if include_statistics is None:
            return cls()

        downsampled_tensor = cls.downsample_tensor(tensor=tensor, sample_percentage=sample_percentage)
        statistics_dict = {}
        for field_name, field_info in cls.model_fields.items():
            if field_name in include_statistics:
                raw_value = field_info.json_schema_extra[FROM_TENSOR_FN](downsampled_tensor)  # type: ignore
                transform_name = include_statistics[field_name] or DEFAULT_TRANSFORM
                transform_fn = TRANSFORM_REGISTRY[transform_name]
                statistics_dict[field_name] = transform_fn(raw_value)

        return cls(**statistics_dict)

    @classmethod
    def from_tensor(cls, tensor: torch.Tensor) -> "TensorStatistics":
        """
        :param tensor: Tensor to build the model from.
        :return: Reconstructed model.
        """

        return cls(
            norm=tensor[0],
            mean=tensor[1],
            median=tensor[2],
            variance=tensor[3],
            tenth_percentile=tensor[4],
            ninetieth_percentile=tensor[5],
            inter_quartile_range=tensor[6],
        )

    def to_list(self, include_statistics: IncludeStatisticsType) -> list[float]:
        """
        :param include_statistics: List of field names to include in the returned list, or a dict mapping
        field names to transforms (only keys are used, transforms already applied during build).
        :return: List of field values.
        """

        if include_statistics is None:
            return []

        return [
            field_value for field_name, field_value in self.model_dump().items() if field_name in include_statistics
        ]

    def to_tensor(self) -> torch.Tensor:
        """
        :return: Tensor with contents of the Pydantic model in a specific order.
        """

        return torch.tensor(
            [
                self.norm,
                self.mean,
                self.median,
                self.variance,
                self.tenth_percentile,
                self.ninetieth_percentile,
                self.inter_quartile_range,
            ]
        )



# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def calculate_adam_beta_gain(adam_beta_one: float, adam_beta_two: float) -> float:
    """
    Utility function expressing relationship between adam parameters and beta gain.
    """
    return (adam_beta_two - adam_beta_one) / (1 - adam_beta_one)


def calculate_adam_beta_two(adam_beta_one: float, adam_beta_gain: float) -> float:
    """
    Utility function inverting the relationship between adam parameters and beta gain.
    """
    return (1 - adam_beta_one) * adam_beta_gain + adam_beta_one



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from abc import abstractmethod
from typing import Any, final

import numpy as np

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class Standardizer:
    """
    Abstract class which defines the interface for environment observation standardizers to use.
    """

    def __init__(self, **kwargs) -> None:
        """
        :param kwargs: Keyword arguments included to allow for any signature in subclasses.
        """

        self._running_statistics: dict[str, Any] = {}

    @final
    def __call__(self, statistics_key: str, value_to_standardize: float) -> float:
        """
        :param statistics_key: Key which these statistics are stored under.
        :param value_to_standardize: New metric value to update statistics with and then standardize.
        :return: Standardized value.
        """

        return self.standardize(statistics_key=statistics_key, value_to_standardize=value_to_standardize)

    @final
    def _get_running_statistics(self, statistics_key: str) -> Any:
        """
        :param statistics_key: Key of the statistics to retrieve from stored statistics.
        :return: Stored statistics.
        """

        if statistics_key not in self._running_statistics:
            self._running_statistics[statistics_key] = self._get_default_statistics()

        return self._running_statistics[statistics_key]

    @abstractmethod
    def _get_default_statistics(self) -> Any:
        """
        :return: Default value of the statistics when they are not already present in the stored statistics.
        """

        raise NotImplementedError

    @abstractmethod
    def _standardize_value(self, statistics_key: str, value_to_standardize: float) -> float:
        """
        :param statistics_key: Key which these statistics are stored under.
        :param value_to_standardize: New metric value to standardize.
        :return: Standardized value.
        """

        raise NotImplementedError

    @abstractmethod
    def _update_running_statistics(self, statistics_key: str, value_to_standardize: float) -> None:
        """
        :param statistics_key: Key which these statistics are stored under.
        :param value_to_standardize: New metric value to update statistics with.
        """

        raise NotImplementedError

    @final
    def standardize(self, statistics_key: str, value_to_standardize: float) -> float:
        """
        :param statistics_key: Key which these statistics are stored under.
        :param value_to_standardize: New metric value to update statistics with and then standardize.
        :return: Standardized value.
        """

        try:
            self._update_running_statistics(statistics_key=statistics_key, value_to_standardize=value_to_standardize)
            standardized = self._standardize_value(
                statistics_key=statistics_key, value_to_standardize=value_to_standardize
            )
        except ArithmeticError:
            standardized = float("nan")

        return standardized

    def reset(self) -> None:
        """
        Empties the stored statistics dictionary.
        """

        self._running_statistics = {}


class EMAZScoreStandardizer(Standardizer):
    """
    Standardizer which standardizes metrics according to Z-scores computed from running means and variances using
    Exponential Moving Average.
    """

    def __init__(
        self,
        minimum_standard_deviation: float = 1e-2,
        standardization_epsilon: float = 1e-4,
        ema_learning_rate: float = 0.4,
        default_mean: float = 0,
        default_variance: float = 1,
        default_count: int = 0,
        **kwargs,
    ) -> None:
        """
        :param minimum_standard_deviation: Minimum value the standard deviation of a statistic that can be used in
        Z-score standardization.
        :param standardization_epsilon: Small value to add to the denominator of the Z-score standardization.
        :param ema_learning_rate: Factor which alters how quickly the EMA adapts to new values.
        :param default_mean: Default mean value for newly added statistics.
        :param default_variance: Default variance value for newly added statistics.
        :param default_count: Default counter value for newly added statistics.
        :param kwargs: Keyword arguments included to allow for any signature in subclasses.
        """

        super().__init__(**kwargs)

        self.minimum_standard_deviation = minimum_standard_deviation
        self.standardization_epsilon = standardization_epsilon

        self.ema_learning_rate = ema_learning_rate

        self.default_mean = default_mean
        self.default_variance = default_variance
        self.default_count = default_count

    def _get_default_statistics(self) -> tuple[float, float, int]:
        """
        :return: Default value of the statistics when they are not already present in the stored statistics.
        """

        return self.default_mean, self.default_variance, self.default_count

    def _standardize_value(self, statistics_key: str, value_to_standardize: float) -> float:
        """
        :param statistics_key: Key which these statistics are stored under.
        :param value_to_standardize: New metric value to standardize.
        :return: Standardized value.
        """

        current_ema, current_ema_variance, _ = self._get_running_statistics(statistics_key=statistics_key)

        standard_deviation = np.sqrt(current_ema_variance)
        standard_deviation = max(standard_deviation, self.minimum_standard_deviation)

        numerator = value_to_standardize - current_ema
        denominator = standard_deviation + self.standardization_epsilon

        standard_value = numerator / denominator
        assert isinstance(standard_value, float)

        return standard_value

    def _update_running_statistics(self, statistics_key: str, value_to_standardize: float) -> None:
        """
        :param statistics_key: Key which these statistics are stored under.
        :param value_to_standardize: New metric value to update statistics with.

        :todo: Vectorise this class using numpy.
        """

        current_ema, current_ema_variance, count = self._get_running_statistics(statistics_key=statistics_key)

        count += 1

        new_ema = self.ema_learning_rate * value_to_standardize + (1 - self.ema_learning_rate) * current_ema

        new_variance = self.ema_learning_rate * (value_to_standardize - new_ema) ** 2
        new_ema_variance = new_variance + (1 - self.ema_learning_rate) * current_ema_variance

        self._running_statistics[statistics_key] = (new_ema, new_ema_variance, count)


# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def get_standardizer(
    standardizer_name: str | None,
    standardizer_arguments: dict[str, Any],
) -> Standardizer | None:
    """
    :param standardizer_name: Name of the standardizer to get.
    :param standardizer_arguments: Arguments to instantiate the standardizer with.
    :return: Instantiated standardizer.
    """

    if standardizer_name is None:
        return None

    elif standardizer_name == EMAZScoreStandardizer.__name__:
        standardizer = EMAZScoreStandardizer

    else:
        raise ValueError(f"Unrecognised standardizer {standardizer_name}!")

    return standardizer(**standardizer_arguments)



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

import asyncio
import datetime
import os
import traceback as traceback_module
from types import TracebackType
from typing import Protocol, TypeAlias

from loguru import logger
from slack_sdk.errors import SlackApiError
from slack_sdk.web.async_client import AsyncWebClient as SlackAsyncClient

from libinephany.utils.asyncio_worker import AsyncioWorker
from libinephany.utils.error_severities import SEVERITY_EMOJIS, SEVERITY_LEVELS, ErrorSeverities

# ======================================================================================================================
#
# TYPE HINTS
#
# ======================================================================================================================

ExceptionEntry: TypeAlias = tuple[Exception, traceback_module.TracebackException, datetime.datetime]
ExceptionPackage: TypeAlias = tuple[Exception, str, datetime.datetime]


class ExceptionSeverityClassifier(Protocol):
    def __call__(self, *, exception: Exception, traceback: str, timestamp: datetime.datetime) -> ErrorSeverities:
        """
        :param exception: Exception to classify the severity of.
        :param traceback: Traceback associated with the given exception.
        :param timestamp: Time the exception occurred.
        :return: Severity of the exception.
        """

        ...


# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class ErrorLogger(AsyncioWorker):

    SLACK_BOT_TOKEN = "SLACK_BOT_TOKEN"
    SLACK_BOT_CHANNEL_ID = "SLACK_BOT_CHANNEL_ID"

    ENCODING = "utf-8"
    CHANNEL_MENTION = "<!channel>"

    def __init__(
        self,
        service_name: str,
        member_ids_to_mention: list[str] | None = None,
        custom_error_message: str | None = None,
        minimum_severity: ErrorSeverities = ErrorSeverities.WARNING,
        error_classifier: ExceptionSeverityClassifier | None = None,
        max_time_accumulating_errors: float = 10.0,
        max_errors_in_queue: int = 5000,
    ) -> None:
        """
        :param service_name: Name of the service deployed on the cloud errors sent to this error logger should be sent
        to.
        :param member_ids_to_mention: List of member IDs to mention in the Slack message. If None, no members will be mentioned.
        :param custom_error_message: Custom error message to include in the Slack message. If None, no custom message will be included.
        :param minimum_severity: Minimum error severity that can be sent to Slack.
        :param error_classifier: Optional callable used to alter how errors severities are classified. If this is None
        all errors are given the 'WARNING' severity.
        :param max_time_accumulating_errors: Maximum amount of time to spend batching identical errors together.
        :param max_errors_in_queue: Maximum length of the error logging queue.
        """

        self._error_queue: asyncio.Queue = asyncio.Queue(maxsize=max_errors_in_queue)

        self._slack_client, self._channel_id = self._get_slack_client()

        self.member_mentions = self._form_member_mentions(member_ids_to_mention=member_ids_to_mention)
        self.custom_error_message = custom_error_message
        self.service_name = service_name
        self.error_classifier = error_classifier if error_classifier is not None else default_error_classifier
        self.max_time_accumulating_errors = max_time_accumulating_errors
        self.minimum_severity = SEVERITY_LEVELS[minimum_severity]

        super().__init__()

    def __enter__(self) -> "ErrorLogger":
        """
        :return: This class instance.
        """

        return self

    def __exit__(
        self, exc_type: type[Exception] | None, exc_val: Exception | None, exc_tb: TracebackType | None
    ) -> bool | None:
        """
        :param exc_type: None or the type of exception that occurred.
        :param exc_val: None or the caught exception instance.
        :param exc_tb: None or the traceback of the exception.
        :return: Whether the error should propagate outwards.
        """

        if exc_val is not None:
            logger.error(f"{self.__class__.__name__} context manager caught an exception of type {exc_type.__name__}.")

            exception_strings = traceback_module.format_exception(exc_type, exc_val, exc_tb)
            formatted_exception = "".join(exception_strings)

            self.put(exception=exc_val, traceback=formatted_exception)

            return True

        return None

    @property
    def active(self) -> bool:
        """
        :return: Boolean indicating whether the worker background thread should be activated.
        """

        return self._slack_client is not None and self._channel_id is not None

    @staticmethod
    def _log_error_to_terminal(exception: Exception) -> None:
        """
        :param exception: Exception to log to the terminal.
        """

        try:
            logger.error(f"Server encountered {exception.__class__.__name__}. Logging traceback to terminal.")
            logger.exception(exception)
        except Exception as e:
            print(
                f"{e.__class__.__name__} occurred when attempting to log exception {exception.__class__.__name__} to "
                f"the terminal!"
            )

    @staticmethod
    def _get_slack_message(
        mention: str,
        notification_emoji: str,
        service_name: str,
        exception: Exception,
        frequency: int,
        frequency_time_window: float,
        custom_error_message: str | None,
    ) -> str:
        """
        :param mention: Mention tag of the entire channel or a particular user ID.
        :param notification_emoji: Emoji to include in the message which indicates the error's severity.
        :param service_name: Name of the service the error comes from.
        :param exception: Exception that occurred.
        :param frequency: How frequently the exception occurred in the given time window.
        :param frequency_time_window: Time between the first and last instance of the exception.
        :param custom_error_message: Custom error message to include in the Slack message. If None, no custom message will be included.
        :return: Formatted Slack message string.
        """

        frequency_line = f"{frequency} in {frequency_time_window}s" if frequency > 1 else f"{frequency}"

        header = f"{notification_emoji} *{service_name} Error* {notification_emoji}\n"

        if mention:
            header += f"• *Alerting*: {mention}\n"

        if custom_error_message is not None:
            header += f"• *Custom Message*: {custom_error_message}\n"

        return (
            f"{header}"
            f"• *Service*: {service_name}\n"
            f"• *Error Type*: {exception.__class__.__name__}\n"
            f"• *Frequency*: {frequency_line}\n"
            f"• *Traceback*: "
        )

    @staticmethod
    def _form_member_mentions(member_ids_to_mention: list[str] | None) -> str:
        """
        :param member_ids_to_mention: List of member IDs to mention in the Slack message.
        :return: String of member mentions.
        """

        if not member_ids_to_mention:
            return ""

        return " ".join([f"<@{member_id}>" for member_id in member_ids_to_mention])

    def _get_traceback_file_name(self, exception: Exception, exception_timestamp: str) -> str:
        """
        :param exception: Exception being sent to Slack.
        :param exception_timestamp: Time the exception occurred.
        :return: Filename the exception should be stored in.
        """

        formatted_service_name = self.service_name.replace(" ", "_")

        return f"{formatted_service_name}__{exception.__class__.__name__}__{exception_timestamp}.txt"

    def _get_slack_client(self) -> tuple[SlackAsyncClient | None, str | None]:
        """
        :return: Tuple of:
            - None or the asynchronous Slack API client.
            - None or the ID of the channel the Slack bot should send messages to.
        """

        api_token = os.environ.get(self.SLACK_BOT_TOKEN, None)
        channel_id = os.environ.get(self.SLACK_BOT_CHANNEL_ID, None)

        if api_token is not None and channel_id is not None:
            return SlackAsyncClient(token=api_token), channel_id

        logger.warning(
            f"No errors will be sent to Slack. Missing either {self.SLACK_BOT_TOKEN} or {self.SLACK_BOT_CHANNEL_ID}."
        )
        return None, None

    def _handle_unexpected_exception(self, exception: Exception, error_traceback: str) -> None:
        """
        :param exception: Exception that occurred and was not handled elsewhere.
        :param error_traceback: Traceback of the exception.
        """

    async def _count_exception_frequency(
        self, exception: Exception, timestamp: datetime.datetime
    ) -> tuple[int, float, ExceptionPackage | None]:
        """
        :param exception: Exception to count the frequency of.
        :param timestamp: Time the exception occurred.
        :return: Tuple of:
            - How frequently the error occurred in the returned time window.
            - Time window which is the time in seconds between the first occurrence of the given error and the last
              occurrence.
            - None or a Tuple containing the next error to handle.
        """

        base_entry = (exception, traceback_module.TracebackException.from_exception(exception), timestamp)

        next_to_process, time_window = None, None
        start_time = datetime.datetime.now(datetime.timezone.utc)
        frequency = 1

        while True:
            try:
                next_exception, next_traceback, next_timestamp = await asyncio.wait_for(
                    self._error_queue.get(), timeout=self.max_time_accumulating_errors
                )
                next_entry = (
                    next_exception,
                    traceback_module.TracebackException.from_exception(next_exception),
                    next_timestamp,
                )

                exceptions_are_similar = _exceptions_are_similar(
                    original_exception_entry=base_entry,
                    next_exception_entry=next_entry,
                    start_time=start_time,
                    max_interval=self.max_time_accumulating_errors,
                )

                if not exceptions_are_similar:
                    next_to_process = (next_exception, next_traceback, next_timestamp)
                    break

                frequency += 1
                time_window = next_timestamp - timestamp

            except (asyncio.QueueEmpty, TimeoutError):
                break

        time_window = 0.0 if time_window is None else round(time_window.total_seconds(), 3)
        return frequency, time_window, next_to_process

    async def _send_error_to_slack(
        self,
        severity: ErrorSeverities,
        exception: Exception,
        traceback: str,
        exception_timestamp: datetime.datetime,
        frequency: int,
        frequency_time_window: float,
    ) -> None:
        """
        :param exception: Exception caught by the API middleware and transferred to the ErrorLogger.
        :param traceback: Traceback associated with the exception.
        :param exception_timestamp: Time the exception occurred.
        :param frequency: How many times the error occurred within a certain timeframe.
        :param frequency_time_window: Length of time between the first and last occurrence of this error.

        :todo: Handle rate limits and retries.
        """

        formatted_timestamp = exception_timestamp.isoformat()
        traceback_filename = self._get_traceback_file_name(exception=exception, exception_timestamp=formatted_timestamp)

        notification_emoji = SEVERITY_EMOJIS[severity]
        # Temporary since the bot is used in training and the API.
        mention = (
            self.member_mentions if not ErrorSeverities.should_mention(severity=severity) else self.CHANNEL_MENTION
        )

        self._log_error_to_terminal(exception=exception)

        try:
            await self._slack_client.files_upload_v2(
                channel=self._channel_id,
                filename=traceback_filename,
                content=traceback.encode(self.ENCODING),
                initial_comment=self._get_slack_message(
                    mention=mention,
                    notification_emoji=notification_emoji,
                    service_name=self.service_name,
                    exception=exception,
                    frequency=frequency,
                    frequency_time_window=frequency_time_window,
                    custom_error_message=self.custom_error_message,
                ),
            )
        except SlackApiError as e:
            logger.exception(e)
            logger.error("Failed to send error to slack!")

    async def _work(
        self,
        previous_result: ExceptionPackage | None,
    ) -> ExceptionPackage | None:
        """
        :param previous_result: Previous result of call to this method.
        :return: Result to convey to this method on the next call.

        Asynchronous worker function which gathers errors and sends them to Slack.
        """

        next_exception: ExceptionPackage | None = None

        if next_exception is None:
            exception, traceback, exception_timestamp = await self._error_queue.get()
        else:
            exception, traceback, exception_timestamp = next_exception

        severity = self.error_classifier(exception=exception, traceback=traceback, timestamp=exception_timestamp)
        severity_level = SEVERITY_LEVELS[severity]

        if severity_level <= self.minimum_severity:
            frequency, frequency_window, next_exception = await self._count_exception_frequency(
                exception=exception, timestamp=exception_timestamp
            )

            await self._send_error_to_slack(
                severity=severity,
                exception=exception,
                traceback=traceback,
                exception_timestamp=exception_timestamp,
                frequency=frequency,
                frequency_time_window=frequency_window,
            )

        return next_exception

    def put(self, exception: Exception, traceback: str) -> None:
        """
        :param exception: Exception to add to the queue.
        :param traceback: Traceback of the exception to add to the queue.
        """

        if self.active:
            exception_timestamp = datetime.datetime.now(datetime.timezone.utc)

            try:
                self._error_queue.put_nowait((exception, traceback, exception_timestamp))

            except asyncio.QueueFull:
                logger.warning(f"{self.__class__.__name__} queue is full! Logging error to terminal.")
                self._log_error_to_terminal(exception=exception)

        else:
            self._log_error_to_terminal(exception=exception)


# ======================================================================================================================
#
# FUNCTIONS
#
# ======================================================================================================================


def _extract_last_frame(traceback: traceback_module.TracebackException) -> None | traceback_module.FrameSummary:
    """
    :param traceback: Traceback exception stack to extract the last frame from.
    :return: None or the last traceback frame.
    """

    return traceback.stack[-1] if traceback.stack else None


def _exceptions_are_similar(
    original_exception_entry: ExceptionEntry,
    next_exception_entry: ExceptionEntry,
    start_time: datetime.datetime,
    max_interval: float,
) -> bool:
    """
    :param original_exception_entry: Tuple of the original exception, it's traceback stack and the time it occurred at.
    :param next_exception_entry: Tuple of the new exception, it's traceback stack and the time it occurred at.
    :param start_time: Time this exception batch begun accumulating.
    :param max_interval: Maximum time between identical exceptions.
    :return: Whether the exceptions should be batched together in the Slack message.
    """

    original_exception, original_traceback, original_timestamp = original_exception_entry
    next_exception, next_traceback, next_timestamp = next_exception_entry

    original_last_frame = _extract_last_frame(traceback=original_traceback)
    next_last_frame = _extract_last_frame(traceback=next_traceback)

    if not original_last_frame or not next_last_frame:
        return False

    return (
        original_last_frame.filename == next_last_frame.filename
        and original_last_frame.lineno == next_last_frame.lineno
        and original_last_frame.name == next_last_frame.name
        and original_exception.__class__.__name__ == next_exception.__class__.__name__
        and abs((next_timestamp - start_time).total_seconds()) <= max_interval
    )


def default_error_classifier(*, exception: Exception, traceback: str, timestamp: datetime.datetime) -> ErrorSeverities:
    """
    :param exception: Exception to classify the severity of.
    :param traceback: Traceback associated with the given exception.
    :param timestamp: Time the exception occurred.
    :return: Severity of the exception.

    :note: This is a default placeholder function not intended for actual use in production.
    """

    logger.debug(
        f"Default error classifier received exception {exception.__class__.__name__} at {timestamp}. Giving"
        f" error severity level '{ErrorSeverities.WARNING.value}'."
    )
    logger.trace(traceback)

    return ErrorSeverities.WARNING



# ======================================================================================================================
#
# IMPORTS
#
# ======================================================================================================================

from enum import Enum

from libinephany.utils.constants import (
    ADAM_BETA_GAIN,
    ADAM_BETA_ONE,
    ADAM_BETA_TWO,
    ADAM_EPS,
    BATCH_SIZE,
    DROPOUT,
    EPOCHS,
    GRAD_NORM_CLIP,
    GRADIENT_ACCUMULATION,
    LEARNING_RATE,
    SAMPLES,
    TOKENS,
    WEIGHT_DECAY,
)

# ======================================================================================================================
#
# CLASSES
#
# ======================================================================================================================


class EnumWithIndices(Enum):

    @classmethod
    def get_index(cls, enum_name: str) -> int:
        """
        :param enum_name: Name of the enum to get the position index for.
        :return: Index position of the given enum in the list of possible schemas.
        :raises ValueError: When given enum does not match any types stored in the enum class.
        """

        as_list = list(cls)  # type: ignore

        for i, action in enumerate(as_list):
            if action.value == enum_name:
                return i

        raise ValueError(f"Unknown enum name {enum_name}.")

    @classmethod
    def get_from_index(cls, index: int) -> "EnumWithIndices":
        """
        :param index: Index of the enum to get.
        :return: Enum corresponding to the given index.
        """

        as_list = list(cls)

        return as_list[index]


class AgentTypes(EnumWithIndices):

    # Active Agents
    LearningRateAgent = LEARNING_RATE
    WeightDecayAgent = WEIGHT_DECAY
    DropoutAgent = DROPOUT
    GradientClippingAgent = GRAD_NORM_CLIP
    AdamBetaOneAgent = ADAM_BETA_ONE
    AdamBetaTwoAgent = ADAM_BETA_TWO
    AdamBetaGainAgent = ADAM_BETA_GAIN
    AdamEpsAgent = ADAM_EPS
    GradientAccumulationAgent = GRADIENT_ACCUMULATION

    # Deprecated or Non-Agent
    BatchSize = BATCH_SIZE
    Epochs = EPOCHS
    Tokens = TOKENS
    Samples = SAMPLES

    @classmethod
    def get_possible_active_agents(cls) -> list["AgentTypes"]:
        """
        :return: List of active agents.
        """

        return [
            cls.LearningRateAgent,
            cls.WeightDecayAgent,
            cls.DropoutAgent,
            cls.GradientClippingAgent,
            cls.AdamBetaOneAgent,
            cls.AdamBetaTwoAgent,
            cls.AdamBetaGainAgent,
            cls.AdamEpsAgent,
            cls.GradientAccumulationAgent,
        ]


class ModelFamilies(EnumWithIndices):

    OLMo = "olmo"
    GPT = "gpt"
    BERT = "bert"


class ModuleTypes(EnumWithIndices):

    Convolutional = "convolutional"
    Attention = "attention"
    Linear = "linear"
    Embedding = "embedding"
    LSTM = "lstm"


class ToyTaskName(EnumWithIndices):

    XOR = "XOR"
    NQM = "NQM"


class ValidationFrequencyType(str, Enum):
    EPOCH = "epoch"
    BATCH = "batch"
